package w2phtml.rdf;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Set;
import java.util.Vector;

import org.apache.jena.ontology.OntClass;
import org.apache.jena.ontology.OntModel;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.vocabulary.*;

import w2phtml.xhtml.XhtmlConfig;
import w2phtml.xhtml.XhtmlDocument;

import org.apache.jena.rdf.model.Property;


public class DocumentStructure {
	private static final String HTML_EXCERPT_PROPERTY = "htmlExcerpt";
	private static final String TOCITEM = "TOCItem";
	private static final String TS = "https://litvinovg.pro/text_structures#";
	private static final String PARSERNAME = "w2phtml"; 
	private static final String TEXT_EXCERPT = "textExcerpt";
	private static final String TOC_LEVEL = "TOCLevel";
	private static final String TABLE_OF_CONTENTS = "TOC";
	private static final String PUBLICATION = "publication";
	private static final String BOOK = "book";
	private static final String JOURNAL = "journal";
	private static final String ENC_ARTICLE = "encArticle";
	private static final String ELENPH_ARTICLE = "elenphArticle";
	private static final String ELENPH_EXCERPT = "elenphExcerpt";
	
	private String docID = "DOC_ID";
	private String documentType = "elenphArticle";
	private String excerptType = "elenphExcerpt";
	
	private OntModel m;
	private OntModel textOntology;
	private OntClass excerptClass;
	private OntClass tocLevelClass;
	private OntClass itemClass;
	private OntClass documentClass;
	private OntClass tocClass;

	private HashMap<String, Resource> tocLevels;
	private HashMap<String, DocumentPart> inputParts;
	
	public DocumentStructure(Vector<XhtmlDocument> files,String fileName, XhtmlConfig config) {
		this.tocLevels = new HashMap<String, Resource>();
		this.inputParts = new HashMap<String, DocumentPart>(); 
		this.m = ModelFactory.createOntologyModel();
		this.textOntology = ModelFactory.createOntologyModel();
		this.tocLevelClass = textOntology.createClass(TS + TOC_LEVEL);
		this.tocClass = textOntology.createClass(TS + TABLE_OF_CONTENTS);
		this.itemClass = textOntology.createClass(TS + TOCITEM);
		this.docID = fileName;
		String publicationType = config.getRDFType();
		setPublicationType(publicationType);
		this.documentClass = textOntology.createClass(TS + documentType);
		this.excerptClass = textOntology.createClass(TS + excerptType);
		for(int i = 0 ; i< files.size();i++) {
			XhtmlDocument inputDoc = files.get(i);
			DocumentPart part = new DocumentPart(inputDoc);
			part.setOrder(Integer.toString(i));
			addPart(part);
		}
		addEmptyParts();
	}
	
	private void setPublicationType(String publicationType) {

		if (publicationType.equals(documentType)) {
			this.documentType = ELENPH_ARTICLE;
			this.excerptType = ELENPH_EXCERPT;
		} else if (publicationType.equals(BOOK)){
			this.documentType = BOOK;
			this.excerptType = TEXT_EXCERPT;
		} else if (publicationType.equals(JOURNAL)){
			this.documentType = JOURNAL;
			this.excerptType = TEXT_EXCERPT;
		} else if (publicationType.equals(ENC_ARTICLE)){
			this.documentType = ENC_ARTICLE;
			this.excerptType = TEXT_EXCERPT;
		}
	}

	private void addEmptyParts() {
		Set<String> paths = inputParts.keySet();
		String[] array = new String[paths.size()];
		paths.toArray(array);
		for (int k = 0; k < array.length;k++) {
			String[] levels = array[k].split(" ");
			for (int i = 0; i < levels.length; i++) {
				if (levels[i].equals("0")) {
					String emptyPath = createEmptyPath(levels, i);
					if (!inputParts.containsKey(emptyPath)) {
						DocumentPart emptyPart = new DocumentPart(emptyPath);
						addPart(emptyPart);
					}
				}
			}	
		}
	}
	
	private String createEmptyPath(String[] levels, int i) {
		StringBuilder emptyPath = new StringBuilder();
		for (int j = 0; j<= i;j++) {
			if (j != 0) {
				emptyPath.append(" ");
			}
			emptyPath.append(levels[j]);
		}
		return emptyPath.toString();
	}
		
	public void printModel(String fileName) {
		File outFile = new File(fileName + ".rdf");
		FileWriter fw = null;
		try {
			outFile.createNewFile();
			fw = new FileWriter(outFile);
			m.write(fw,"RDF/XML-ABBREV");
			
		} catch (IOException e) {
			System.out.println("File couldn't be created");
			e.printStackTrace();
		} finally {
			try {
				fw.close();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
	}
	public void printModel(OutputStream os) {
		m.write(os,"RDF/XML-ABBREV");
	}
	
	private void addPart(DocumentPart docExcerpt) {
		inputParts.put(docExcerpt.getPath(), docExcerpt);
	}
	
	private Resource createExcerpt(DocumentPart docExcerpt) {
		String excerptName = TS + excerptType + "/" + PARSERNAME + "_" + docID + docExcerpt.getSafePath();
		Resource excerpt = m.createIndividual(excerptName, excerptClass);
		if (!docExcerpt.getBody().isEmpty()) {
			Property htmlExcerpt = m.createProperty(TS + HTML_EXCERPT_PROPERTY);
			excerpt.addLiteral(htmlExcerpt, docExcerpt.getBody());	
		}
		return excerpt;
	}
	
	private void createTOCLevel(DocumentPart docPart) {
		String levelName = TS + TOC_LEVEL + "/" + PARSERNAME + "_" + docID + docPart.getSafePath();
		Resource level = m.createIndividual(levelName,tocLevelClass);
		level.addProperty( RDFS.label, docPart.getName());
		tocLevels.put(docPart.getPath(), level);
		attachExcerpt(docPart, level);		
	}	
	
	private void createDocument(DocumentPart docPart) {
		String documentURI = TS + documentType + "/" + PARSERNAME + "_" + docID ;
		Resource document = m.createResource(documentURI, documentClass);
		document.addProperty( RDFS.label, docPart.getName());
		String tocURI = TS + TABLE_OF_CONTENTS + "/" + PARSERNAME + "_" + docID ;

		Resource toc = m.createResource(tocURI, tocClass);
		toc.addProperty( RDFS.label, docPart.getName());

		Property hasTOC = m.createProperty(TS + "hasTOC");
		document.addProperty(hasTOC, toc);
				
		addMetadataProperties(document, docPart.getMetadata());
		tocLevels.put(docPart.getPath(), toc);
		attachExcerpt(docPart, toc);	
	}
	
	private void addMetadataProperties(Resource resource, HashMap<String, Set<String>> metadata) {
		Set<String> names = metadata.keySet();
		for (String name : names) {
			Set<String> values = metadata.get(name);
			if (values != null) {
				for (String value : values) {
					addMetadata(resource,name,value);
				}
			}
		}
	}
	private void addMetadata(Resource resource, String name, String value) {
		if (isNotBlacklisted(name)) {
			name = convertName(name);
			if (isDefinedInOntology(resource,name)) {
				if (name.equals("keywords")) {
					addKeywords(resource, name, value);
				} else {
					Property property = m.createProperty(TS + name);
					resource.addProperty( property, value);
				}
			}
		}
	}
	private void addKeywords(Resource resource,String name, String value) {
		String[] keywords = value.split(",");
		for (int i = 0; i < keywords.length; i++) {
			String keyword = keywords[i].trim();
			if (!keyword.isEmpty()) {
				Property property = m.createProperty(TS + name);
				resource.addProperty( property, keyword);
			}
		}
	}
	
	private boolean isDefinedInOntology(Resource resource, String name) {
		String nameSpace = resource.getNameSpace();
		if (nameSpace.contains(TS + excerptType)) {
			if (name.equals("author") ||
					name.equals("bibliography") ||
					name.equals("keywords") ||
					name.equals("works") ||
					name.equals("affiliation")
					) {
				return true;
			}
		} else 
		if (nameSpace.contains(TS + documentType)) {
			if (name.equals("doi") ||
					name.equals("firstPublication") ||
					//name.equals("yearAndMonth") ||
					name.equals("year") ||
					name.equals("issue")
					) {
				return true;
			}
		}
		System.out.println("Skip " + name + " for " + resource.getNameSpace());
		return false;
	}
	private String convertName(String name) {
		if (name.equals("Affiliation")){
			return "affiliation";
		} else 
		if (name.equals("DOI")){
			return "doi";
		} else
		if (name.equals("1st-edition")){
			return "firstPublication";
		} else
		if (name.equals("year")){
			return "yearAndMonth";
		} else
		if (name.equals("year.short")){
				return "year";
		}else
		return name;
	}
	private boolean isNotBlacklisted(String name) {
		if (name.equals("Filename") ||
				name.equals("Section") ||
				name.equals("dc.Title") ||
				name.equals("subtitle") ||
				name.equals("dc.Identifier")
				) {
				return false;
		}
	
		return true;
	}
	private void createTOCItem(DocumentPart docPart) {
		String tocItemName = TS + TOCITEM + "/" + PARSERNAME + "_" + docID + docPart.getSafePath();
		Resource tocItem = m.createIndividual(tocItemName,itemClass);
		tocItem.addProperty( RDFS.label, docPart.getName());
		Property pointsTo = m.createProperty(TS + "pointsTo");
		Property itemNumber = m.createProperty(TS + "itemNumber");
		Property hasTOCItem = m.createProperty(TS + "hasTOCItem");	
		tocItem.addLiteral(itemNumber, docPart.getNumber());
		m.add(tocItem, pointsTo, tocLevels.get(docPart.getPath()));
		if (!docPart.getPath().isEmpty()) {
			Resource parent = tocLevels.get(docPart.getParentPath());
			m.add(parent, hasTOCItem, tocItem);
		}
	
	}
	
	private void attachExcerpt(DocumentPart docPart, Resource element) {
		if (docPart.isEmpty()) {
			return;
		}
		if (docPart.getBody().isEmpty()) {
			return;
		}
		Resource excerpt = createExcerpt(docPart);
		excerpt.addProperty( RDFS.label, docPart.getName());
		Property hasText = m.createProperty(TS + "hasText");
		element.addProperty(hasText, excerpt);
		if (!docPart.isMasterPart()) {
			addMetadataProperties(excerpt,docPart.getMetadata());
		}
	}
	
	public void createTree() {
		createElements();
		createTOCItems();
	}
	
	private void createTOCItems() {
		Set<String> paths = inputParts.keySet();
		for (String path : paths) {
			DocumentPart part = inputParts.get(path);
			if (!part.getNumber().equals("")) {
				createTOCItem(part);	
			}
		}		
	}

	private void createElements() {
		Set<String> paths = inputParts.keySet();
		for (String path : paths) {
			DocumentPart part = inputParts.get(path);
			if (part.getPath().isEmpty()) {
				createDocument(part);
			} else {
				createTOCLevel(part);
			}
		}
	}
	public void applyMetadata(Metadata metadata) {
		for (DocumentPart part: inputParts.values()) {
			 metadata.apply(part);
		}
	}

}
